// Author: A.Voss@FH-Aachen.de
//
// control ctor and op= implementation
// tags: default delete

#include <iostream>
#include <cstdlib>
#include <vector>
#include <memory>
using namespace std;

#include "../../include/tools.h"

#pragma GCC diagnostic ignored "-Wunused-but-set-variable"
#pragma GCC optimize "-fno-elide-constructors"

class A { 
public:
    A() = default;
    A(double) { }         // if this ctor exists, no default is created
    A(const A&) = delete; // no copy ctor
    
    A& operator=(const A&) = default;    // and default assignment
    
    // old style: make it private or avoid implementation
    // A& operator=(const A&);
};

class C {
public:
    // most C++11 features
    
    // create a std ctor
    C() = default;
    
    // use class ctor C(string) 
    C(const char * pc) : C(string(pc))  { tools_log(); cout << "   C(pc) '" << pc << "'" << endl; }
    
    // is used by C(char*)
    C(const string & s)                 { tools_log(); cout << "   C(s)  '" << s << "'" << endl; }
    
    // cannot be called implicitly
    explicit C(int n)                   { tools_log(); cout << "   C(n)  '" << n << "'" << endl; }
    
    // copy ctor deleted
    C(const C &) = delete;
    
    // move
    C(C &&)                             { tools_log(); cout << "   C(&&) '?'" << endl; }
    
    // standard dstr
    ~C()    = default;
    
    // C& operator=(const C&) = default; //{ return *this; }
};

int main()
{
    tools_log(); cout << "A objects" << endl << endl;
    A a1;
    A a2(2.3);
    // A a3(a2); // deleted
    a1=a2;       // default

    
    tools_log(); cout << "std ctor" << endl;
    C c1;

    // whats the difference?
    tools_log(); cout << "ctor str" << endl;
    C c2(string("c2"));
    tools_log(); cout << "ctor pc" << endl;
    C c3("c3");
    tools_log(); cout << "ctor pc" << endl;
    C c4 = "c4";
    
    tools_log(); cout << "ctor n" << endl;
    C c5(42);
    //C c5=42;    // does not work, why?

    tools_log(); cout << "ctor copy" << endl;
    //C c6(c5)    // does not work, why?
    C c6=move(c5);
    
    // c1 = c2; // does not work, why?
 
    // remarks
    // c4: note, a move ctor is called    
    // c5: implicit conversion is not allowed ("explicit")    
    // c6: copy ctor deleted, but move works
    //
    // some operators are deleted implicitly if others are defined,
    // here op= is deleted because C(&&) is user-defined (of course,
    // you may still define your own)
    
    return EXIT_SUCCESS;
}

